import math
from itertools import product
from pathlib import Path
from time import time

import dill
import pandas as pd
import pytest

from tests.framework.results.helpers import (
    FAMILIARS,
    HARRY_POTTER_CONFIG,
    POWER_LEVEL_GROUP_LABELS,
    STUDENT_HOUSES,
    Hogwarts,
    HogwartsResultsStratifier,
    HousePointsObserver,
    NoStratificationsQuidditchWinsObserver,
    QuidditchWinsObserver,
)
from tests.helpers import Listener, MockComponentA, MockComponentB
from vivarium import Component
from vivarium.framework.artifact import ArtifactInterface, ArtifactManager
from vivarium.framework.components import (
    ComponentConfigError,
    ComponentInterface,
    ComponentManager,
    OrderedComponentSet,
)
from vivarium.framework.engine import Builder
from vivarium.framework.engine import SimulationContext as SimulationContext_
from vivarium.framework.event import EventInterface, EventManager
from vivarium.framework.lifecycle import LifeCycleInterface, LifeCycleManager
from vivarium.framework.logging import LoggingInterface, LoggingManager
from vivarium.framework.lookup import LookupTableInterface, LookupTableManager
from vivarium.framework.population import PopulationInterface, PopulationManager
from vivarium.framework.randomness import RandomnessInterface, RandomnessManager
from vivarium.framework.resource import ResourceInterface, ResourceManager
from vivarium.framework.results import VALUE_COLUMN, ResultsInterface, ResultsManager
from vivarium.framework.time import DateTimeClock, TimeInterface
from vivarium.framework.values import ValuesInterface, ValuesManager


def is_same_object_method(m1, m2):
    return m1.__func__ is m2.__func__ and m1.__self__ is m2.__self__


@pytest.fixture()
def SimulationContext():
    yield SimulationContext_
    SimulationContext_._clear_context_cache()


@pytest.fixture
def components():
    return [
        MockComponentA("gretchen", "whimsy"),
        Listener("listener"),
        MockComponentB("spoon", "antelope", "23"),
    ]


@pytest.fixture
def log(mocker):
    return mocker.patch("vivarium.framework.logging.manager.loguru.logger")


def test_simulation_with_non_components(SimulationContext, components: list[Component]):
    class NonComponent:
        def __init__(self):
            self.name = "non_component"

    with pytest.raises(ComponentConfigError):
        SimulationContext(components=components + [NonComponent()])


def test_SimulationContext_get_sim_name(SimulationContext):
    assert SimulationContext._created_simulation_contexts == set()

    assert SimulationContext._get_context_name(None) == "simulation_1"
    assert SimulationContext._get_context_name("foo") == "foo"

    assert SimulationContext._created_simulation_contexts == {"simulation_1", "foo"}


def test_SimulationContext_init_default(SimulationContext, components):
    sim = SimulationContext(components=components)

    assert isinstance(sim._logging, LoggingManager)
    assert isinstance(sim._lifecycle, LifeCycleManager)
    assert isinstance(sim._component_manager, ComponentManager)
    assert isinstance(sim._clock, DateTimeClock)
    assert isinstance(sim._values, ValuesManager)
    assert isinstance(sim._events, EventManager)
    assert isinstance(sim._population, PopulationManager)
    assert isinstance(sim._resource, ResourceManager)
    assert isinstance(sim._results, ResultsManager)
    assert isinstance(sim._tables, LookupTableManager)
    assert isinstance(sim._randomness, RandomnessManager)
    assert isinstance(sim._data, ArtifactManager)

    assert isinstance(sim._builder, Builder)
    assert sim._builder.configuration is sim.configuration

    assert isinstance(sim._builder.logging, LoggingInterface)
    assert sim._builder.logging._manager is sim._logging
    assert isinstance(sim._builder.lookup, LookupTableInterface)
    assert sim._builder.lookup._manager is sim._tables
    assert isinstance(sim._builder.value, ValuesInterface)
    assert sim._builder.value._manager is sim._values
    assert isinstance(sim._builder.event, EventInterface)
    assert sim._builder.event._manager is sim._events
    assert isinstance(sim._builder.population, PopulationInterface)
    assert sim._builder.population._manager is sim._population
    assert isinstance(sim._builder.resources, ResourceInterface)
    assert sim._builder.resources._manager is sim._resource
    assert isinstance(sim._builder.results, ResultsInterface)
    assert sim._builder.results._manager is sim._results
    assert isinstance(sim._builder.randomness, RandomnessInterface)
    assert sim._builder.randomness._manager is sim._randomness
    assert isinstance(sim._builder.time, TimeInterface)
    assert sim._builder.time._manager is sim._clock
    assert isinstance(sim._builder.components, ComponentInterface)
    assert sim._builder.components._manager is sim._component_manager
    assert isinstance(sim._builder.lifecycle, LifeCycleInterface)
    assert sim._builder.lifecycle._manager is sim._lifecycle
    assert isinstance(sim._builder.data, ArtifactInterface)
    assert sim._builder.data._manager is sim._data

    # Ordering matters.
    managers = [
        sim._logging,
        sim._lifecycle,
        sim._resource,
        sim._values,
        sim._population,
        sim._clock,
        sim._randomness,
        sim._events,
        sim._tables,
        sim._data,
        sim._results,
    ]
    assert sim._component_manager._managers == OrderedComponentSet(*managers)
    unpacked_components = []
    for c in components:
        unpacked_components.append(c)
        if hasattr(c, "sub_components"):
            unpacked_components.extend(c.sub_components)
    assert list(sim._component_manager._components) == unpacked_components


def test_SimulationContext_name_management(SimulationContext):
    assert SimulationContext._created_simulation_contexts == set()

    sim1 = SimulationContext()
    assert sim1._name == "simulation_1"
    assert SimulationContext._created_simulation_contexts == {"simulation_1"}

    sim2 = SimulationContext(sim_name="foo")
    assert sim2._name == "foo"
    assert SimulationContext._created_simulation_contexts == {"simulation_1", "foo"}

    sim3 = SimulationContext()
    assert sim3._name == "simulation_3"
    assert SimulationContext._created_simulation_contexts == {
        "simulation_1",
        "foo",
        "simulation_3",
    }


def test_SimulationContext_run_simulation(SimulationContext, mocker):
    sim = SimulationContext()

    expected_calls = [
        "setup",
        "initialize_simulants",
        "run",
        "finalize",
        "report",
    ]

    # Mock the methods called by sim.run()
    mock = mocker.MagicMock()
    for call in expected_calls:
        mock.attach_mock(
            mocker.patch(f"vivarium.framework.engine.SimulationContext.{call}"), call
        )

    sim.run_simulation()

    # Assert the calls are each made exactly once and in the correct order
    # NOTE: mock.mock_calls is a list like [call.setup(), call.initialize_simulants(), ...]
    actual_calls = [str(call).split("call.")[1].split("()")[0] for call in mock.mock_calls]
    assert actual_calls == expected_calls


def test_SimulationContext_setup_default(SimulationContext, base_config, components):
    sim = SimulationContext(base_config, components)
    listener = [c for c in components if "listener" in c.args][0]
    assert not listener.post_setup_called
    sim.setup()

    unpacked_components = []
    for c in components:
        unpacked_components.append(c)
        if hasattr(c, "sub_components"):
            unpacked_components.extend(c.sub_components)

    for a, b in zip(sim._component_manager._components, unpacked_components):
        assert type(a) == type(b)
        if hasattr(a, "args"):
            assert a.args == b.args

    assert is_same_object_method(sim.simulant_creator, sim._population._create_simulants)
    assert sim.time_step_events == [
        "time_step__prepare",
        "time_step",
        "time_step__cleanup",
        "collect_metrics",
    ]
    for k in sim.time_step_emitters.keys():
        assert is_same_object_method(
            sim.time_step_emitters[k], sim._events._event_types[k].emit
        )

    assert is_same_object_method(
        sim.end_emitter, sim._events._event_types["simulation_end"].emit
    )

    assert listener.post_setup_called


def test_SimulationContext_initialize_simulants(SimulationContext, base_config, components):
    sim = SimulationContext(base_config, components)
    sim.setup()
    pop_size = sim.configuration.population.population_size
    current_time = sim._clock.time
    assert sim._population.get_population(True).empty
    sim.initialize_simulants()
    pop = sim._population.get_population(True)
    assert len(pop) == pop_size
    assert sim._clock.time == current_time


def test_SimulationContext_step(SimulationContext, log, base_config, components):
    sim = SimulationContext(base_config, components)
    sim.setup()
    sim.initialize_simulants()

    current_time = sim._clock.time
    step_size = sim._clock.step_size

    listener = [c for c in components if "listener" in c.args][0]

    assert not listener.time_step_prepare_called
    assert not listener.time_step_called
    assert not listener.time_step_cleanup_called
    assert not listener.collect_metrics_called

    sim.step()

    assert log.debug.called_once_with(current_time)
    assert listener.time_step_prepare_called
    assert listener.time_step_called
    assert listener.time_step_cleanup_called
    assert listener.collect_metrics_called

    assert sim._clock.time == current_time + step_size


def test_SimulationContext_finalize(SimulationContext, base_config, components):
    sim = SimulationContext(base_config, components)
    listener = [c for c in components if "listener" in c.args][0]
    sim.setup()
    sim.initialize_simulants()
    sim.step()
    assert not listener.simulation_end_called
    sim.finalize()
    assert listener.simulation_end_called


def test_get_results(SimulationContext, base_config):
    """Test that get_results returns expected values. This does NOT test for
    correct formatting.
    """
    components = [
        Hogwarts(),
        HousePointsObserver(),
        NoStratificationsQuidditchWinsObserver(),
        QuidditchWinsObserver(),
        HogwartsResultsStratifier(),
    ]
    sim = SimulationContext(base_config, components, configuration=HARRY_POTTER_CONFIG)
    sim.run_simulation()
    for measure, results in sim.get_results().items():
        raw_results = sim._results._raw_results[measure].sort_index()
        assert results.set_index(raw_results.index.names)[[VALUE_COLUMN]].equals(raw_results)


def test_SimulationContext_report_no_write_warning(SimulationContext, base_config, caplog):
    components = [
        Hogwarts(),
        HousePointsObserver(),
        NoStratificationsQuidditchWinsObserver(),
        QuidditchWinsObserver(),
        HogwartsResultsStratifier(),
    ]
    sim = SimulationContext(base_config, components, configuration=HARRY_POTTER_CONFIG)
    sim.run_simulation()
    assert "No results directory set; results are not written to disk." in caplog.text
    results = sim.get_results()
    assert set(results) == set(
        ["house_points", "quidditch_wins", "no_stratifications_quidditch_wins"]
    )
    assert all([isinstance(df, pd.DataFrame) for df in results.values()])


def test_SimulationContext_report_write(SimulationContext, base_config, components, tmpdir):
    """Test that the written results match get_results"""
    results_root = Path(tmpdir)
    configuration = {"output_data": {"results_directory": str(results_root)}}
    configuration.update(HARRY_POTTER_CONFIG)
    components = [
        Hogwarts(),
        HousePointsObserver(),
        NoStratificationsQuidditchWinsObserver(),
        QuidditchWinsObserver(),
        HogwartsResultsStratifier(),
    ]
    sim = SimulationContext(base_config, components, configuration)
    sim.run_simulation()
    # Check for expected results written
    results_list = [file for file in results_root.rglob("*.parquet")]
    assert set([file.name for file in results_list]) == set(
        [
            "house_points.parquet",
            "quidditch_wins.parquet",
            "no_stratifications_quidditch_wins.parquet",
        ]
    )

    # Check that written results match get_results method
    for measure in ["house_points", "quidditch_wins", "no_stratifications_quidditch_wins"]:
        results = sim.get_results()[measure]
        written_results = pd.read_parquet(results_root / f"{measure}.parquet")
        assert results.equals(written_results)


def test_SimulationContext_write_backup(mocker, SimulationContext, tmpdir):
    # TODO MIC-5216: Remove mocks when we can use dill in pytest.
    mocker.patch("vivarium.framework.engine.dill.dump")
    mocker.patch("vivarium.framework.engine.dill.load", return_value=SimulationContext())
    sim = SimulationContext()
    backup_path = tmpdir / "backup.pkl"
    sim.write_backup(backup_path)
    assert backup_path.exists()
    with open(backup_path, "rb") as f:
        sim_backup = dill.load(f)
    assert isinstance(sim_backup, SimulationContext)


def test_SimulationContext_run_with_backup(mocker, SimulationContext, base_config, tmpdir):
    mocker.patch("vivarium.framework.engine.SimulationContext.write_backup")
    original_time = time()

    def time_generator():
        current_time = original_time
        while True:
            yield current_time
            current_time += 5

    mocker.patch("vivarium.framework.engine.time", side_effect=time_generator())
    components = [
        Hogwarts(),
        HousePointsObserver(),
        NoStratificationsQuidditchWinsObserver(),
        QuidditchWinsObserver(),
        HogwartsResultsStratifier(),
    ]
    sim = SimulationContext(base_config, components, configuration=HARRY_POTTER_CONFIG)
    backup_path = tmpdir / "backup.pkl"
    sim.setup()
    sim.initialize_simulants()
    sim.run(backup_path=backup_path, backup_freq=5)
    assert sim.write_backup.call_count == _get_num_steps(sim)


def test_get_results_formatting(SimulationContext, base_config):
    """Test formatted results are as expected"""
    components = [
        Hogwarts(),
        HousePointsObserver(),
        NoStratificationsQuidditchWinsObserver(),
        QuidditchWinsObserver(),
        HogwartsResultsStratifier(),
    ]
    sim = SimulationContext(base_config, components, configuration=HARRY_POTTER_CONFIG)
    sim.run_simulation()
    num_steps = _get_num_steps(sim)

    house_points = sim.get_results()["house_points"]
    quidditch_wins = sim.get_results()["quidditch_wins"]
    no_stratifications_quidditch_wins = sim.get_results()["no_stratifications_quidditch_wins"]

    # Check that each dataset includes the entire cartesian product of stratifications
    # (or, when no stratifications, just a single "all" row)
    assert set(zip(house_points["student_house"], house_points["power_level_group"])) == set(
        product(STUDENT_HOUSES, POWER_LEVEL_GROUP_LABELS)
    )
    assert set(quidditch_wins["familiar"]) == set(FAMILIARS)
    assert no_stratifications_quidditch_wins.shape[0] == 1
    assert (no_stratifications_quidditch_wins["stratification"] == "all").all()
    assert set(quidditch_wins.columns).difference(
        set(no_stratifications_quidditch_wins.columns)
    ) == set(["familiar"])

    # Set up filters for groups that scored points
    house_points_filter = (house_points["student_house"] == "gryffindor") & (
        house_points["power_level_group"].isin(["low", "very high"])
    )
    quidditch_wins_filter = quidditch_wins["familiar"] == "banana_slug"
    no_strats_quidditch_wins_filter = (
        no_stratifications_quidditch_wins["stratification"] == "all"
    )
    for measure, filter in [
        ("house_points", house_points_filter),
        ("quidditch_wins", quidditch_wins_filter),
        ("no_stratifications_quidditch_wins", no_strats_quidditch_wins_filter),
    ]:
        # Check columns
        df = eval(measure)
        # Check that metrics col matches name of dataset
        assert (df["measure"] == measure).all()
        # We do enforce a col order, but most importantly ensure VALUE_COLUMN is at the end
        assert df.columns[-1] == VALUE_COLUMN
        # Check values
        # Check that all values are 0 except for expected groups
        assert (df.loc[filter, VALUE_COLUMN] != 0).all()
        assert (df.loc[~filter, VALUE_COLUMN] == 0).all()
        # Check that expected groups' values are a multiple of the number of steps
        assert (df.loc[filter, VALUE_COLUMN] % num_steps == 0).all()


####################
# HELPER FUNCTIONS #
####################
def _convert_to_datetime(date_dict: dict[str, int]) -> pd.Timestamp:
    return pd.to_datetime(
        "-".join([str(val) for val in date_dict.values()]), format="%Y-%m-%d"
    )


def _get_num_steps(sim: SimulationContext) -> int:
    time_dict = sim.configuration.time.to_dict()
    end_date = _convert_to_datetime(time_dict["end"])
    start_date = _convert_to_datetime(time_dict["start"])
    num_steps = math.ceil((end_date - start_date).days / time_dict["step_size"])
    return num_steps



from vivarium.framework.components.manager import (
    ComponentConfigError,
    ComponentInterface,
    ComponentManager,
    OrderedComponentSet,
)
from vivarium.framework.components.parser import ComponentConfigurationParser




from __future__ import annotations
from abc import ABC, abstractmethod
from typing import Any

import numpy as np
import pandas as pd

from vivarium import Component
from vivarium.framework.engine import Builder


class Force(Component, ABC):
    ##############
    # Properties #
    ##############
    @property
    def configuration_defaults(self) -> dict[str, Any]:
        return {
            self.__class__.__name__.lower(): {
                "max_force": 0.03,
            },
        }

    columns_required = []

    #####################
    # Lifecycle methods #
    #####################

    def setup(self, builder: Builder) -> None:
        self.config = builder.configuration[self.__class__.__name__.lower()]
        self.max_speed = builder.configuration.movement.max_speed

        self.neighbors = builder.value.get_value("neighbors")

        builder.value.register_value_modifier(
            "acceleration",
            modifier=self.apply_force,
            required_resources=self.columns_required + [self.neighbors],
        )

    ##################################
    # Pipeline sources and modifiers #
    ##################################

    def apply_force(self, index: pd.Index[int], acceleration: pd.DataFrame) -> pd.DataFrame:
        neighbors = self.neighbors(index)
        pop = self.population_view.get(index)
        pairs = self._get_pairs(neighbors, pop)

        raw_force = self.calculate_force(pairs)
        force = self._normalize_and_limit_force(
            force=raw_force,
            velocity=pop[["vx", "vy"]],
            max_force=self.config.max_force,
            max_speed=self.max_speed,
        )

        acceleration.loc[force.index, ["x", "y"]] += force[["x", "y"]]
        return acceleration

    ##################
    # Helper methods #
    ##################

    @abstractmethod
    def calculate_force(self, neighbors: pd.DataFrame) -> pd.DataFrame:
        pass

    def _get_pairs(self, neighbors: pd.Series[int], pop: pd.DataFrame) -> pd.DataFrame:
        pairs = (
            pop.join(neighbors.rename("neighbors"))
            .reset_index()
            .explode("neighbors")
            .merge(
                pop.reset_index(),
                left_on="neighbors",
                right_index=True,
                suffixes=("_self", "_other"),
            )
        )
        pairs["distance_x"] = pairs.x_other - pairs.x_self
        pairs["distance_y"] = pairs.y_other - pairs.y_self
        pairs["distance"] = np.sqrt(pairs.distance_x**2 + pairs.distance_y**2)

        return pairs

    def _normalize_and_limit_force(
        self,
        force: pd.DataFrame,
        velocity: pd.DataFrame,
        max_force: float,
        max_speed: float,
    ) -> pd.DataFrame:
        normalization_factor = np.where(
            (force.x != 0) | (force.y != 0),
            max_speed / self._magnitude(force),
            1.0,
        )
        force["x"] *= normalization_factor
        force["y"] *= normalization_factor
        force["x"] -= velocity.loc[force.index, "vx"]
        force["y"] -= velocity.loc[force.index, "vy"]
        magnitude = self._magnitude(force)
        limit_scaling_factor = np.where(
            magnitude > max_force,
            max_force / magnitude,
            1.0,
        )
        force["x"] *= limit_scaling_factor
        force["y"] *= limit_scaling_factor
        return force[["x", "y"]]

    def _magnitude(self, df: pd.DataFrame) -> pd.Series[float]:
        return pd.Series(np.sqrt(np.square(df.x) + np.square(df.y)), dtype=float)


class Separation(Force):
    """Push boids apart when they get too close."""

    configuration_defaults = {
        "separation": {
            "distance": 30,
            "max_force": 0.03,
        },
    }

    def calculate_force(self, neighbors: pd.DataFrame) -> pd.DataFrame:
        # Push boids apart when they get too close
        separation_neighbors = neighbors[neighbors.distance < self.config.distance].copy()
        force_scaling_factor = np.where(
            separation_neighbors.distance > 0,
            ((-1 / separation_neighbors.distance) / separation_neighbors.distance),
            1.0,
        )
        separation_neighbors["force_x"] = (
            separation_neighbors["distance_x"] * force_scaling_factor
        )
        separation_neighbors["force_y"] = (
            separation_neighbors["distance_y"] * force_scaling_factor
        )

        force: pd.DataFrame = (
            separation_neighbors.groupby("index_self")[["force_x", "force_y"]]
            .sum()
            .rename(columns=lambda c: c.replace("force_", ""))
        )

        return force


class Cohesion(Force):
    """Push boids together."""

    def calculate_force(self, pairs: pd.DataFrame) -> pd.DataFrame:
        return (
            pairs.groupby("index_self")[["distance_x", "distance_y"]]
            .sum()
            .rename(columns=lambda c: c.replace("distance_", ""))
        )


class Alignment(Force):
    """Push boids toward where others are going."""

    def calculate_force(self, pairs: pd.DataFrame) -> pd.DataFrame:
        return (
            pairs.groupby("index_self")[["vx_other", "vy_other"]]
            .sum()
            .rename(columns=lambda c: c.replace("v", "").replace("_other", ""))
        )



import itertools

import numpy as np
import pandas as pd

from vivarium import Component
from vivarium.framework.engine import Builder
from vivarium.framework.event import Event
from vivarium.framework.population import SimulantData
from vivarium.framework.results import VALUE_COLUMN
from vivarium.framework.results.observer import Observer

NAME = "hogwarts_house"
NAME_COLUMNS = ["first_name", "last_name"]
HOUSE_CATEGORIES = ["hufflepuff", "ravenclaw", "slytherin", "gryffindor"]
STUDENT_TABLE = pd.DataFrame(
    np.array([["harry", "potter"], ["severus", "snape"], ["luna", "lovegood"]]),
    columns=NAME_COLUMNS,
)
STUDENT_HOUSES = pd.Series(["gryffindor", "slytherin", "ravenclaw"])

BIN_BINNED_COLUMN = "silly_bin"
BIN_SOURCE = "silly_level"
BIN_LABELS = ["meh", "somewhat", "very", "extra"]
BIN_SILLY_BIN_EDGES = [0, 20, 40, 60, 90]

COL_NAMES = ["house", "familiar", "power_level", "tracked"]
FAMILIARS = ["owl", "cat", "gecko", "banana_slug", "unladen_swallow"]
POWER_LEVELS = [20, 40, 60, 80]
POWER_LEVEL_BIN_EDGES = [0, 25, 50, 75, 100]
POWER_LEVEL_GROUP_LABELS = ["low", "medium", "high", "very high"]
TRACKED_STATUSES = [True, False]
RECORDS = list(itertools.product(HOUSE_CATEGORIES, FAMILIARS, POWER_LEVELS, TRACKED_STATUSES))
BASE_POPULATION = pd.DataFrame(data=RECORDS, columns=COL_NAMES)

HARRY_POTTER_CONFIG = {
    "time": {
        "start": {"year": 2024, "month": 4, "day": 22},
        "end": {"year": 2029, "month": 4, "day": 22},
        "step_size": 365,  # Days
    },
    "stratification": {
        "default": ["student_house", "power_level_group"],
    },
}

RNG = np.random.default_rng(42)


##################
# Helper classes #
##################


class Hogwarts(Component):
    @property
    def columns_created(self) -> list[str]:
        return [
            "student_id",
            "student_house",
            "familiar",
            "power_level",
            "house_points",
            "quidditch_wins",
            "exam_score",
            "spell_power",
            "potion_power",
        ]

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        size = len(pop_data.index)
        initialization_data = pd.DataFrame(
            {
                "student_id": list(range(size)),
                "student_house": RNG.choice(STUDENT_HOUSES, size=size),
                "familiar": RNG.choice(FAMILIARS, size=size),
                "power_level": RNG.choice([lvl for lvl in POWER_LEVELS], size=size),
                "house_points": 0,
                "quidditch_wins": 0,
                "exam_score": 0.0,
            },
            index=pop_data.index,
        )
        # Assume power level components are evenly split
        initialization_data["spell_power"] = initialization_data["power_level"] / 2
        initialization_data["potion_power"] = initialization_data["power_level"] / 2
        self.population_view.update(initialization_data)

    def on_time_step(self, pop_data: Event) -> None:
        update = self.population_view.get(pop_data.index)
        update["house_points"] = 0
        update["quidditch_wins"] = 0
        # House points are stratified by 'student_house' and 'power_level_group'.
        # Let's have each wizard of gryffindor and of power level 20 or 80
        # gain a point on each time step.
        update.loc[
            (update["student_house"] == "gryffindor")
            & (update["power_level"].isin([20, 80])),
            "house_points",
        ] = 1
        # Quidditch wins are stratified by 'familiar'.
        # Let's have each wizard with a banana slug familiar gain a point
        # on each time step.
        update.loc[update["familiar"] == "banana_slug", "quidditch_wins"] = 1
        # Update everyones test score to increase by 10 points per time step
        update["exam_score"] += 10.0

        self.population_view.update(update)


class HousePointsObserver(Observer):
    """Observer that is stratified by multiple columns (the defaults,
    'student_house' and 'power_level_group')
    """

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(
            name="house_points",
            aggregator_sources=["house_points"],
            aggregator=sum,
            requires_columns=[
                "house_points",
            ],
            results_formatter=results_formatter,
        )


class FullyFilteredHousePointsObserver(Observer):
    """Same as `HousePointsObserver but with a filter that leaves no simulants"""

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(
            name="house_points",
            pop_filter="tracked==True & power_level=='one billion'",
            aggregator_sources=["house_points"],
            aggregator=sum,
            requires_columns=[
                "house_points",
            ],
        )


class QuidditchWinsObserver(Observer):
    """Observer that is stratified by a single column ('familiar')"""

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(
            name="quidditch_wins",
            aggregator_sources=["quidditch_wins"],
            aggregator=sum,
            excluded_stratifications=["student_house", "power_level_group"],
            additional_stratifications=["familiar"],
            requires_columns=[
                "quidditch_wins",
            ],
            results_formatter=results_formatter,
        )


class NoStratificationsQuidditchWinsObserver(Observer):
    """Same as above but no stratifications at all"""

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(
            name="no_stratifications_quidditch_wins",
            aggregator_sources=["quidditch_wins"],
            aggregator=sum,
            excluded_stratifications=["student_house", "power_level_group"],
            requires_columns=[
                "quidditch_wins",
            ],
            results_formatter=results_formatter,
        )


class MagicalAttributesObserver(Observer):
    """Observer whose aggregator returns a pd.Series instead of a float (which in
    turn results in a dataframe with multiple columns instead of just one
    'value' column)
    """

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(
            name="magical_attributes",
            aggregator=self._get_magical_attributes,
            excluded_stratifications=["student_house"],
            results_formatter=lambda *_: None,
        )

    def _get_magical_attributes(self, _: pd.DataFrame) -> pd.Series:
        """Increase each level by one per time step"""
        return pd.Series([1.0, 1.0], ["spell_power", "potion_power"])


class ExamScoreObserver(Observer):
    """Observer that is not stratified and exam scores"""

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_concatenating_observation(
            name="exam_score",
            requires_columns=["student_id", "student_house", "exam_score"],
        )


class CatBombObserver(Observer):
    """Observer that counts the number of feral cats per house"""

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_stratified_observation(
            name="cat_bomb",
            pop_filter="familiar=='cat' and tracked==True",
            requires_columns=["familiar"],
            results_updater=self.update_cats,
            excluded_stratifications=["power_level_group"],
            aggregator_sources=["student_house"],
            aggregator=len,
        )

    def update_cats(self, existing_df, new_df):
        no_cats_mask = existing_df["value"] == 0
        updated_df = existing_df
        updated_df.loc[no_cats_mask, "value"] = new_df["value"]
        updated_df.loc[~no_cats_mask, "value"] *= new_df["value"]
        return updated_df


class ValedictorianObserver(Observer):
    """Observer that records the valedictorian at each time step. All students
    have the same exam scores and so the valedictorian is chosen randomly.
    """

    def __init__(self):
        super().__init__()
        self.valedictorians = []

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_unstratified_observation(
            name="valedictorian",
            requires_columns=["event_time", "student_id", "exam_score"],
            results_gatherer=self.choose_valedictorian,
            results_updater=self.update_valedictorian,
        )

    def choose_valedictorian(self, df):
        eligible_students = df.loc[~df["student_id"].isin(self.valedictorians), "student_id"]
        valedictorian = RNG.choice(eligible_students)
        self.valedictorians.append(valedictorian)
        return df[df["student_id"] == valedictorian]

    def update_valedictorian(self, _existing_df, new_df):
        return new_df


class HogwartsResultsStratifier(Component):
    def setup(self, builder: Builder) -> None:
        builder.results.register_stratification(
            name="student_house",
            categories=list(STUDENT_HOUSES),
            requires_columns=["student_house"],
        )
        builder.results.register_stratification(
            name="familiar", categories=FAMILIARS, requires_columns=["familiar"]
        )
        builder.results.register_binned_stratification(
            "power_level",
            "power_level_group",
            POWER_LEVEL_BIN_EDGES,
            POWER_LEVEL_GROUP_LABELS,
        )


##################
# Helper methods #
##################


def results_formatter(
    measure: str,
    results: pd.DataFrame,
) -> pd.DataFrame:
    """An test use case of an observer's report method that writes a DataFrame to a CSV file."""
    # Add extra cols
    results["measure"] = measure
    # Sort the columns such that the stratifications (index) are first
    # and VALUE_COLUMN is last and sort the rows by the stratifications.
    other_cols = [c for c in results.columns if c != VALUE_COLUMN]
    return results[other_cols + [VALUE_COLUMN]].sort_index().reset_index()


def sorting_hat_vectorized(state_table: pd.DataFrame) -> pd.Series:
    sorted_series = state_table.apply(sorting_hat_serial, axis=1)
    return sorted_series


def sorting_hat_serial(simulant_row: pd.Series) -> str:
    first_name = simulant_row[0]
    last_name = simulant_row[1]
    if first_name == "harry":
        return "gryffindor"
    if first_name == "luna":
        return "ravenclaw"
    if last_name == "snape":
        return "slytherin"
    return "hufflepuff"


def sorting_hat_bad_mapping(simulant_row: pd.Series) -> str:
    # Return something not in CATEGORIES
    return "pancakes"


def verify_stratification_added(
    stratification_list, name, sources, categories, excluded_categories, mapper, is_vectorized
):
    """Verify that a Stratification object is in `stratification_list`"""
    matching_stratification_found = False
    for stratification in stratification_list:  # noqa
        # big equality check
        if (
            stratification.name == name
            and sorted(stratification.categories)
            == sorted([cat for cat in categories if cat not in excluded_categories])
            and sorted(stratification.excluded_categories) == sorted(excluded_categories)
            and stratification.mapper == mapper
            and stratification.is_vectorized == is_vectorized
            and sorted(stratification.sources) == sorted(sources)
        ):
            matching_stratification_found = True
            break
    return matching_stratification_found


# Mock for get_value call for Pipelines, returns a str instead of a Pipeline
def mock_get_value(self, name: str):
    if not isinstance(name, str):
        raise TypeError("Passed a non-string type to mock get_value(), check your pipelines.")
    return name



from __future__ import annotations

from collections.abc import Callable

import numpy as np
import pandas as pd
import pytest
from layered_config_tree import LayeredConfigTree

from tests.helpers import ColumnCreator
from vivarium import InteractiveContext
from vivarium.framework.configuration import build_simulation_configuration
from vivarium.framework.engine import Builder
from vivarium.framework.population import SimulantData
from vivarium.framework.resource import Resource
from vivarium.framework.state_machine import Machine, State, Transition
from vivarium.types import ClockTime, LookupTableData


def test_initialize_allowing_self_transition() -> None:
    self_transitions = State("self-transitions", allow_self_transition=True)
    no_self_transitions = State("no-self-transitions", allow_self_transition=False)
    undefined_self_transitions = State("self-transitions")

    assert self_transitions.transition_set.allow_null_transition
    assert not no_self_transitions.transition_set.allow_null_transition
    assert not undefined_self_transitions.transition_set.allow_null_transition


def test_initialize_with_initial_state() -> None:
    start_state = State("start")
    other_state = State("other")
    machine = Machine("state", states=[start_state, other_state], initial_state=start_state)
    simulation = InteractiveContext(components=[machine])
    assert simulation.get_population()["state"].unique() == ["start"]


@pytest.mark.parametrize("weights_type", ["artifact", "callable", "scalar"])
def test_initialize_with_scalar_initialization_weights(
    base_config: LayeredConfigTree, weights_type: str
) -> None:
    state_weights = {"state_a.weights": 0.2, "state_b.weights": 0.8}

    def mock_load(key: str) -> float:
        return state_weights.get(key)

    base_config.update(
        {"population": {"population_size": 10000}, "randomness": {"key_columns": []}}
    )

    def initialization_weights(
        key: str,
    ) -> LookupTableData | str | Callable[[Builder], LookupTableData]:
        return {
            "artifact": key,
            "callable": lambda _: state_weights[key],
            "scalar": state_weights[key],
        }[weights_type]

    state_a = State("a", initialization_weights=initialization_weights("state_a.weights"))
    state_b = State("b", initialization_weights=initialization_weights("state_b.weights"))
    machine = Machine("state", states=[state_a, state_b])
    simulation = InteractiveContext(
        components=[machine], configuration=base_config, setup=False
    )
    simulation._builder.data.load = mock_load
    simulation.setup()

    state = simulation.get_population()["state"]
    assert np.all(simulation.get_population().state != "start")
    assert round((state == "a").mean(), 1) == 0.2
    assert round((state == "b").mean(), 1) == 0.8


@pytest.mark.parametrize("weights_type", ["artifact", "callable", "dataframe"])
def test_initialize_with_array_initialization_weights(weights_type: str) -> None:
    state_weights = {
        "state_a.weights": pd.DataFrame(
            {"test_column_1": [0, 1, 2], "value": [0.2, 0.7, 0.4]}
        ),
        "state_b.weights": pd.DataFrame(
            {"test_column_1": [0, 1, 2], "value": [0.8, 0.3, 0.6]}
        ),
    }

    def mock_load(key: str) -> pd.DataFrame:
        return state_weights.get(key)

    config = build_simulation_configuration()
    config.update(
        {"population": {"population_size": 10000}, "randomness ": {"key_columns": []}}
    )

    class TestMachine(Machine):
        @property
        def initialization_requirements(self) -> list[str | Resource]:
            # FIXME - MIC-5408: We shouldn't need to specify the columns in the
            #  lookup tables here, since the component can't know what will be
            #  specified by the states or the configuration.
            return ["test_column_1"]

    def initialization_weights(
        key: str,
    ) -> LookupTableData | str | Callable[[Builder], LookupTableData]:
        return {
            "artifact": key,
            "callable": lambda _: state_weights[key],
            "dataframe": state_weights[key],
        }[weights_type]

    state_a = State("a", initialization_weights=initialization_weights("state_a.weights"))
    state_b = State("b", initialization_weights=initialization_weights("state_b.weights"))
    machine = TestMachine("state", states=[state_a, state_b])
    simulation = InteractiveContext(
        components=[machine, ColumnCreator()], configuration=config, setup=False
    )
    simulation._builder.data.load = mock_load
    simulation.setup()

    pop = simulation.get_population()[["state", "test_column_1"]]
    state_a_weights = state_weights["state_a.weights"]
    state_b_weights = state_weights["state_b.weights"]
    for i in range(3):
        pop_i_state = pop.loc[pop["test_column_1"] == i, "state"]
        assert round((pop_i_state == "a").mean(), 1) == state_a_weights.loc[i, "value"]
        assert round((pop_i_state == "b").mean(), 1) == state_b_weights.loc[i, "value"]


def test_error_if_initialize_with_both_initial_state_and_initialization_weights() -> None:
    start_state = State("start")
    other_state = State("other", initialization_weights=lambda _: 0.8)
    with pytest.raises(ValueError, match="Cannot specify both"):
        Machine("state", states=[start_state, other_state], initial_state=start_state)


def test_error_if_initialize_with_neither_initial_state_nor_initialization_weights() -> None:
    with pytest.raises(ValueError, match="Must specify either"):
        Machine("state", states=[State("a"), State("b")])


@pytest.mark.parametrize("population_size", [1, 100])
@pytest.mark.parametrize("use_transition_arg", [True, False])
def test_transition(
    base_config: LayeredConfigTree, population_size: int, use_transition_arg: bool
) -> None:
    base_config.update(
        {
            "population": {"population_size": population_size},
            "randomness": {"key_columns": []},
        }
    )
    done_state = State("done")
    start_state = State("start")
    if use_transition_arg:
        start_state.add_transition(Transition(start_state, done_state))
    else:
        start_state.add_transition(output_state=done_state)
    machine = Machine("state", states=[start_state, done_state], initial_state=start_state)

    simulation = InteractiveContext(components=[machine], configuration=base_config)
    assert np.all(simulation.get_population().state == "start")
    simulation.step()
    assert np.all(simulation.get_population().state == "done")


def test_no_null_transition(base_config: LayeredConfigTree) -> None:
    base_config.update(
        {"population": {"population_size": 10000}, "randomness": {"key_columns": []}}
    )
    a_state = State("a")
    b_state = State("b")
    start_state = State("start")
    start_state.add_transition(
        output_state=a_state, probability_function=lambda index: pd.Series(0.4, index=index)
    )
    start_state.add_transition(
        output_state=b_state, probability_function=lambda index: pd.Series(0.6, index=index)
    )
    machine = Machine(
        "state", states=[start_state, a_state, b_state], initial_state=start_state
    )

    simulation = InteractiveContext(components=[machine], configuration=base_config)
    assert np.all(simulation.get_population().state == "start")

    simulation.step()

    state = simulation.get_population()["state"]
    assert np.all(simulation.get_population().state != "start")
    assert round((state == "a").mean(), 1) == 0.4
    assert round((state == "b").mean(), 1) == 0.6


def test_null_transition(base_config: LayeredConfigTree) -> None:
    base_config.update(
        {"population": {"population_size": 10000}, "randomness": {"key_columns": []}}
    )
    a_state = State("a")
    start_state = State("start", allow_self_transition=True)
    start_state.add_transition(
        output_state=a_state, probability_function=lambda index: pd.Series(0.4, index=index)
    )

    machine = Machine("state", states=[start_state, a_state], initial_state=start_state)

    simulation = InteractiveContext(components=[machine], configuration=base_config)
    simulation.step()
    state = simulation.get_population()["state"]
    assert round((state == "a").mean(), 1) == 0.4


def test_side_effects() -> None:
    class CountingState(State):
        @property
        def columns_created(self) -> list[str]:
            return ["count"]

        def on_initialize_simulants(self, pop_data: SimulantData) -> None:
            self.population_view.update(pd.Series(0, index=pop_data.index, name="count"))

        def transition_side_effect(self, index: pd.Index[int], _: ClockTime) -> None:
            pop = self.population_view.get(index)
            self.population_view.update(pop["count"] + 1)

    counting_state = CountingState("counting")
    start_state = State("start")
    start_state.add_transition(output_state=counting_state)
    counting_state.add_transition(output_state=start_state)

    machine = Machine(
        "state", states=[start_state, counting_state], initial_state=start_state
    )
    simulation = InteractiveContext(components=[machine])
    assert np.all(simulation.get_population()["count"] == 0)

    # transitioning to counting state
    simulation.step()
    assert np.all(simulation.get_population()["count"] == 1)

    # transitioning back to start state
    simulation.step()
    assert np.all(simulation.get_population()["count"] == 1)

    # transitioning to counting state again
    simulation.step()
    assert np.all(simulation.get_population()["count"] == 2)



"""
=======
Manager
=======

A base Manager class to be used to create manager for use in ``vivarium``
simulations.

"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from vivarium.framework.engine import Builder
    from vivarium.framework.population import SimulantData


class Manager(ABC):
    CONFIGURATION_DEFAULTS: dict[str, Any] = {}
    """A dictionary containing the defaults for any configurations managed by this
    manager. An empty dictionary indicates no managed configurations.

    """

    ##############
    # Properties #
    ##############

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    def configuration_defaults(self) -> dict[str, Any]:
        """Provides a dictionary containing the defaults for any configurations
        managed by this manager.

        These default values will be stored at the `component_configs` layer of the
        simulation's LayeredConfigTree.
        """
        return self.CONFIGURATION_DEFAULTS

    @property
    def columns_created(self) -> list[str]:
        """Provides names of columns created by the manager."""
        return []

    #####################
    # Lifecycle methods #
    #####################

    def setup(self, builder: Builder) -> None:
        """Defines custom actions this manager needs to run during the setup
        lifecycle phase.

        This method is intended to be overridden by subclasses to perform any
        necessary setup operations specific to the manager. By default, it
        does nothing.

        Parameters
        ----------
        builder
            The builder object used to set up the manager.
        """
        pass

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        """
        Method that vivarium will run during simulant initialization.

        This method is intended to be overridden by subclasses if there are
        operations they need to perform specifically during the simulant
        initialization phase.

        Parameters
        ----------
        pop_data : SimulantData
            The data associated with the simulants being initialized.
        """
        pass


class Interface:
    """An interface class to be used to manage different systems for a simulation  in ``vivarium``"""

    pass



from collections.abc import Callable
from typing import Any, Protocol


class ValueCombiner(Protocol):
    def __call__(
        self, value: Any, mutator: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> Any:
        ...


def replace_combiner(
    value: Any, mutator: Callable[..., Any], *args: Any, **kwargs: Any
) -> Any:
    """Replace the previous pipeline output with the output of the mutator.

    This is the default combiner.

    Parameters
    ----------
    value
        The value from the previous step in the pipeline.
    mutator
        A callable that takes in all arguments that the pipeline source takes
        in plus an additional last positional argument for the value from
        the previous stage in the pipeline.
    args, kwargs
        The same args and kwargs provided during the invocation of the
        pipeline.

    Returns
    -------
        A modified version of the input value.
    """
    expanded_args = list(args) + [value]
    return mutator(*expanded_args, **kwargs)


def list_combiner(
    value: list[Any], mutator: Callable[..., Any], *args: Any, **kwargs: Any
) -> list[Any]:
    """Aggregates source and mutator output into a list.

    This combiner is meant to be used with a post-processor that does some
    kind of reduce operation like summing all values in the list.

    Parameters
    ----------
    value
        A list of all values provided by the source and prior mutators in the
        pipeline.
    mutator
        A callable that returns some portion of this pipeline's final value.
    args, kwargs
        The same args and kwargs provided during the invocation of the
        pipeline.

    Returns
    -------
        The input list with new mutator portion of the pipeline value
        appended to it.
    """
    value.append(mutator(*args, **kwargs))
    return value



from __future__ import annotations
import numpy as np
import pandas as pd

from vivarium.framework.event import Event
from vivarium import Component
from vivarium.framework.engine import Builder
from vivarium.framework.population import SimulantData


class Movement(Component):
    ##############
    # Properties #
    ##############
    configuration_defaults = {
        "field": {
            "width": 1000,
            "height": 1000,
        },
        "movement": {
            "max_speed": 2,
        },
    }

    columns_created = ["x", "vx", "y", "vy"]

    #####################
    # Lifecycle methods #
    #####################

    def setup(self, builder: Builder) -> None:
        self.config = builder.configuration

        self.acceleration = builder.value.register_value_producer(
            "acceleration", source=self.base_acceleration
        )

    ##################################
    # Pipeline sources and modifiers #
    ##################################

    def base_acceleration(self, index: pd.Index[int]) -> pd.DataFrame:
        return pd.DataFrame(0.0, columns=["x", "y"], index=index)

    ########################
    # Event-driven methods #
    ########################

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        count = len(pop_data.index)
        # Start randomly distributed, with random velocities
        new_population = pd.DataFrame(
            {
                "x": self.config.field.width * np.random.random(count),
                "y": self.config.field.height * np.random.random(count),
                "vx": ((2 * np.random.random(count)) - 1) * self.config.movement.max_speed,
                "vy": ((2 * np.random.random(count)) - 1) * self.config.movement.max_speed,
            },
            index=pop_data.index,
        )
        self.population_view.update(new_population)

    def on_time_step(self, event: Event) -> None:
        pop = self.population_view.get(event.index)

        acceleration = self.acceleration(event.index)

        # Accelerate and limit velocity
        pop[["vx", "vy"]] += acceleration.rename(columns=lambda c: f"v{c}")
        speed = np.sqrt(np.square(pop.vx) + np.square(pop.vy))
        velocity_scaling_factor = np.where(
            speed > self.config.movement.max_speed,
            self.config.movement.max_speed / speed,
            1.0,
        )
        pop["vx"] *= velocity_scaling_factor
        pop["vy"] *= velocity_scaling_factor

        # Move according to velocity
        pop["x"] += pop.vx
        pop["y"] += pop.vy

        # Loop around boundaries
        pop["x"] = pop.x % self.config.field.width
        pop["y"] = pop.y % self.config.field.height

        self.population_view.update(pop)



from __future__ import annotations

from collections.abc import Callable
from datetime import datetime
from typing import Any

import pytest
import pytest_mock

from tests.helpers import ColumnCreator, ColumnCreatorAndRequirer, ColumnRequirer
from vivarium import Component
from vivarium.framework.population import SimulantData
from vivarium.framework.randomness import RandomnessStream
from vivarium.framework.randomness.index_map import IndexMap
from vivarium.framework.resource import ResourceManager
from vivarium.framework.resource.exceptions import ResourceError
from vivarium.framework.resource.resource import Column, NullResource
from vivarium.framework.values import Pipeline, ValueModifier, ValueSource


@pytest.fixture
def manager(mocker: pytest_mock.MockFixture) -> ResourceManager:
    manager = ResourceManager()
    manager.logger = mocker.Mock()
    return manager


@pytest.fixture
def resource_producers() -> dict[int, ResourceProducer]:
    return {i: ResourceProducer(f"test_{i}") for i in range(5)}


@pytest.fixture
def manager_with_resources(
    manager: ResourceManager, resource_producers: dict[int, ResourceProducer]
) -> ResourceManager:
    stream = RandomnessStream(
        "B", lambda: datetime.now(), 1, IndexMap(), resource_producers[1]
    )
    pipeline = Pipeline("C", resource_producers[2])

    manager.add_resources(resource_producers[3], ["D"], [stream, pipeline])
    manager.add_resources(stream.component, [stream], ["A"])
    manager.add_resources(pipeline.component, [pipeline], ["A"])
    manager.add_resources(resource_producers[0], ["A"], [])
    manager.add_resources(resource_producers[4], [], [stream])
    return manager


@pytest.fixture
def randomness_stream() -> RandomnessStream:
    return RandomnessStream(
        "stream.1", lambda: datetime.now(), 1, IndexMap(), component=ColumnCreator()
    )


class ResourceProducer(Component):
    @property
    def name(self) -> str:
        return self._name

    def __init__(self, name: str):
        super().__init__()
        self._name = name

    def on_initialize_simulants(self, _simulant_data: SimulantData) -> None:
        pass


@pytest.mark.parametrize(
    "resource_class, init_args, type_string, is_initializer",
    [
        (Pipeline, ["foo"], "value", False),
        (ValueSource, [Pipeline("foo"), lambda: 1], "value_source", False),
        (ValueModifier, [Pipeline("foo"), lambda: 1], "value_modifier", False),
        (Column, ["foo"], "column", True),
        (NullResource, [1], "null", True),
    ],
    ids=lambda x: [x.__name__ if isinstance(x, type) else x],
)
def test_resource_manager_get_resource_group(
    resource_class: type,
    init_args: list[Any],
    type_string: str,
    is_initializer: bool,
    manager: ResourceManager,
) -> None:
    component = ColumnCreator()

    group = manager._get_resource_group(
        component, [resource_class(*init_args, component=component)], []
    )

    assert group.type == type_string
    assert group.names == [r.resource_id for r in group.resources.values()]
    assert not group.dependencies
    assert group.is_initialized == is_initializer
    assert group.initializer == component.on_initialize_simulants


def test_resource_manager_get_resource_group_null(manager: ResourceManager) -> None:
    component_1 = ColumnCreator()
    component_2 = ColumnCreatorAndRequirer()

    group_1 = manager._get_resource_group(component_1, [], [])
    group_2 = manager._get_resource_group(component_2, [], [])

    assert group_1.type == "null"
    assert group_1.names == ["null.0"]
    assert group_1.initializer == component_1.on_initialize_simulants
    assert not group_1.dependencies

    assert group_2.type == "null"
    assert group_2.names == ["null.1"]
    assert group_2.initializer == component_2.on_initialize_simulants
    assert not group_2.dependencies


def test_add_resource_wrong_component(manager: ResourceManager) -> None:
    resource = Pipeline("foo", ColumnCreatorAndRequirer())
    error_message = "All initialized resources must have the component 'column_creator'."
    with pytest.raises(ResourceError, match=error_message):
        manager.add_resources(ColumnCreator(), [resource], [])


@pytest.mark.parametrize(
    "resource_type, resource_creator",
    [
        ("column", lambda name, component: name),
        ("value", lambda name, component: Pipeline(name, component)),
    ],
)
def test_resource_manager_add_same_resource_twice(
    resource_type: str,
    resource_creator: Callable[[str, Component], Any],
    manager: ResourceManager,
) -> None:
    c1 = ColumnCreator()
    c2 = ColumnCreatorAndRequirer()
    r1 = [resource_creator(str(i), c1) for i in range(5)]
    r2 = [resource_creator(str(i), c2) for i in range(5, 10)] + [resource_creator("1", c2)]

    manager.add_resources(c1, r1, [])
    error_message = (
        f"Component '{c2.name}' is attempting to register resource"
        f" '{resource_type}.1' but it is already registered by '{c1.name}'."
    )
    with pytest.raises(ResourceError, match=error_message):
        manager.add_resources(c2, r2, [])


def test_resource_manager_sorted_nodes_two_node_cycle(
    manager: ResourceManager, randomness_stream: RandomnessStream
) -> None:
    manager.add_resources(ColumnCreatorAndRequirer(), ["c_1"], [randomness_stream])
    manager.add_resources(randomness_stream.component, [randomness_stream], ["c_1"])

    with pytest.raises(ResourceError, match="The resource pool contains at least one cycle"):
        _ = manager.sorted_nodes


def test_resource_manager_sorted_nodes_three_node_cycle(
    manager: ResourceManager, randomness_stream: RandomnessStream
) -> None:
    pipeline = Pipeline("some_pipeline", ColumnRequirer())

    manager.add_resources(ColumnCreatorAndRequirer(), ["c_1"], [randomness_stream])
    manager.add_resources(pipeline.component, [pipeline], ["c_1"])
    manager.add_resources(randomness_stream.component, [randomness_stream], [pipeline])

    with pytest.raises(ResourceError, match="The resource pool contains at least one cycle"):
        _ = manager.sorted_nodes


def test_resource_manager_sorted_nodes_large_cycle(manager: ResourceManager) -> None:
    component = ColumnCreator()
    for i in range(10):
        manager.add_resources(component, [f"c_{i}"], [f"c_{i % 10}"])

    with pytest.raises(ResourceError, match="cycle"):
        _ = manager.sorted_nodes


def test_large_dependency_chain(manager: ResourceManager) -> None:
    component = ColumnCreator()
    for i in range(9, 0, -1):
        manager.add_resources(component, [f"c_{i}"], [f"c_{i - 1}"])
    manager.add_resources(component, ["c_0"], [])

    for i, resource in enumerate(manager.sorted_nodes):
        assert str(resource) == f"(column.c_{i})"


def test_resource_manager_sorted_nodes_acyclic(
    manager_with_resources: ResourceManager,
) -> None:

    n = [str(node) for node in manager_with_resources.sorted_nodes]

    assert n.index("(column.A)") < n.index("(stream.B)")
    assert n.index("(column.A)") < n.index("(value.C)")
    assert n.index("(column.A)") < n.index("(column.D)")

    assert n.index("(stream.B)") < n.index("(column.D)")
    assert n.index("(value.C)") < n.index("(column.D)")

    assert n.index("(stream.B)") < n.index(f"(null.0)")


def test_get_population_initializers(
    manager_with_resources: ResourceManager, resource_producers: dict[int, ResourceProducer]
) -> None:
    initializers = manager_with_resources.get_population_initializers()

    assert len(initializers) == 3
    assert initializers[0] == resource_producers[0].on_initialize_simulants
    assert resource_producers[3].on_initialize_simulants in initializers
    assert resource_producers[4].on_initialize_simulants in initializers



import itertools
from typing import Sequence

import numpy as np
import pandas as pd
import pytest
from layered_config_tree import LayeredConfigTree
from pytest_mock import MockerFixture

from vivarium import InteractiveContext
from vivarium.framework.lookup import (
    LookupTableInterface,
    LookupTableManager,
    validate_build_table_parameters,
)
from vivarium.framework.lookup.table import InterpolatedTable
from vivarium.testing_utilities import TestPopulation, build_table
from vivarium.types import LookupTableData


@pytest.mark.skip(reason="only order 0 interpolation currently supported")
def test_interpolated_tables(base_config: LayeredConfigTree) -> None:
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    years_df = build_table(
        lambda x: x[0],
        parameter_columns={
            "year": (year_start, year_end),
            "age": (0, 125),
        },
    )
    ages_df = build_table(
        lambda x: x[0],
        parameter_columns={
            "year": (year_start, year_end),
            "age": (0, 125),
        },
    )
    one_d_age_df = ages_df.copy().drop_duplicates()
    base_config.update(
        {"population": {"population_size": 10000}, "interpolation": {"order": 1}}
    )  # the results we're checking later assume interp order 1

    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    years = manager.build_table(
        years_df,
        key_columns=("sex",),
        parameter_columns=(
            "age",
            "year",
        ),
        value_columns=(),
    )
    ages = manager.build_table(
        ages_df,
        key_columns=("sex",),
        parameter_columns=(
            "age",
            "year",
        ),
        value_columns=(),
    )
    one_d_age = manager.build_table(
        one_d_age_df, key_columns=("sex",), parameter_columns=("age",), value_columns=()
    )

    pop = simulation.get_population(untracked=True)
    result_years = years(pop.index)
    result_ages = ages(pop.index)
    result_ages_1d = one_d_age(pop.index)

    fractional_year = simulation._clock.time.year  # type: ignore [union-attr]
    fractional_year += simulation._clock.time.timetuple().tm_yday / 365.25  # type: ignore [union-attr]

    assert np.allclose(result_years, fractional_year)
    assert np.allclose(result_ages, pop.age)
    assert np.allclose(result_ages_1d, pop.age)

    simulation._clock._clock_time += pd.Timedelta(30.5 * 125, unit="D")  # type: ignore [operator]
    simulation._population._population.age += 125 / 12  # type: ignore [union-attr]

    result_years = years(pop.index)
    result_ages = ages(pop.index)
    result_ages_1d = one_d_age(pop.index)

    fractional_year = simulation._clock.time.year  # type: ignore [union-attr]
    fractional_year += simulation._clock.time.timetuple().tm_yday / 365.25  # type: ignore [union-attr]

    assert np.allclose(result_years, fractional_year)
    assert np.allclose(result_ages, pop.age)
    assert np.allclose(result_ages_1d, pop.age)


@pytest.mark.skip(reason="only order 0 interpolation currently supported")
def test_interpolated_tables_without_uninterpolated_columns(
    base_config: LayeredConfigTree,
) -> None:
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    years_df = build_table(
        lambda x: x[0],
        parameter_columns={
            "year": (year_start, year_end),
            "age": (0, 125),
        },
    )
    del years_df["sex"]
    years_df = years_df.drop_duplicates()
    base_config.update(
        {"population": {"population_size": 10000}, "interpolation": {"order": 1}}
    )  # the results we're checking later assume interp order 1

    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    years = manager.build_table(
        years_df,
        key_columns=(),
        parameter_columns=(
            "year",
            "age",
        ),
        value_columns=(),
    )

    result_years = years(simulation.get_population().index)

    fractional_year = simulation._clock.time.year  # type: ignore [union-attr]
    fractional_year += simulation._clock.time.timetuple().tm_yday / 365.25  # type: ignore [union-attr]

    assert np.allclose(result_years, fractional_year)

    simulation._clock._clock_time += pd.Timedelta(30.5 * 125, unit="D")  # type: ignore [operator]

    result_years = years(simulation.get_population().index)

    fractional_year = simulation._clock.time.year  # type: ignore [union-attr]
    fractional_year += simulation._clock.time.timetuple().tm_yday / 365.25  # type: ignore [union-attr]

    assert np.allclose(result_years, fractional_year)


def test_interpolated_tables__exact_values_at_input_points(
    base_config: LayeredConfigTree,
) -> None:
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    years_df = build_table(
        lambda x: x[0],
        parameter_columns={
            "year": (year_start, year_end),
            "age": (0, 125),
        },
    )

    input_years = years_df.year_start.unique()
    base_config.update({"population": {"population_size": 10000}})

    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    years = manager._build_table(
        years_df, key_columns=["sex"], parameter_columns=["age", "year"], value_columns=()
    )

    for year in input_years:
        simulation._clock._clock_time = pd.Timestamp(year, 1, 1)
        assert np.allclose(
            years(simulation.get_population().index), simulation._clock.time.year + 1 / 365  # type: ignore [union-attr]
        )


def test_interpolated_tables__only_categorical_parameters(
    base_config: LayeredConfigTree,
) -> None:
    sexes = ["Female", "Male"]
    locations = ["USA", "Canada", "Mexico"]
    combinations = enumerate(itertools.product(sexes, locations))
    input_data = pd.DataFrame(
        [
            {"sex": sex, "location": location, "some_value": i**2}
            for i, (sex, location) in combinations
        ]
    )

    base_config.update({"population": {"population_size": 10000}})

    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    lookup_table = manager._build_table(
        input_data, key_columns=["sex", "location"], parameter_columns=(), value_columns=()
    )

    population = simulation.get_population()[["sex", "location"]]
    output_data = lookup_table(population.index)

    for i, (sex, location) in combinations:
        sub_table_mask = (output_data["sex"] == sex) & output_data["location"] == location
        assert (output_data.loc[sub_table_mask, "some_value"] == i**2).all()


@pytest.mark.parametrize("data", [(1, 2), [1, 2]])
def test_lookup_table_scalar_from_list(
    base_config: LayeredConfigTree, data: Sequence[int]
) -> None:
    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    table = manager._build_table(
        data, key_columns=(), parameter_columns=(), value_columns=["a", "b"]  # type: ignore [arg-type]
    )(simulation.get_population().index)

    assert isinstance(table, pd.DataFrame)
    assert table.columns.values.tolist() == ["a", "b"]
    assert np.all(table.a == 1)
    assert np.all(table.b == 2)


def test_lookup_table_scalar_from_single_value(base_config: LayeredConfigTree) -> None:
    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    table = manager._build_table(
        1, key_columns=(), parameter_columns=(), value_columns=["a"]
    )(simulation.get_population().index)
    assert isinstance(table, pd.Series)
    assert np.all(table == 1)


def test_invalid_data_type_build_table(base_config: LayeredConfigTree) -> None:
    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    with pytest.raises(TypeError):
        manager._build_table("break", key_columns=(), parameter_columns=(), value_columns=())  # type: ignore [arg-type]


def test_lookup_table_interpolated_return_types(base_config: LayeredConfigTree) -> None:
    year_start = base_config.time.start.year
    year_end = base_config.time.end.year
    data = build_table(
        lambda x: x[0],
        parameter_columns={
            "year": (year_start, year_end),
            "age": (0, 125),
        },
    )

    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    table = manager._build_table(
        data, key_columns=["sex"], parameter_columns=["age", "year"], value_columns=()
    )(simulation.get_population().index)
    # make sure a single value column is returned as a series
    assert isinstance(table, pd.Series)

    # now add a second value column to make sure the result is a df
    data["value2"] = data.value
    table = manager._build_table(
        data, key_columns=["sex"], parameter_columns=["age", "year"], value_columns=()
    )(simulation.get_population().index)

    assert isinstance(table, pd.DataFrame)


@pytest.mark.parametrize(
    "data", [None, pd.DataFrame(), pd.DataFrame(columns=["a", "b", "c"]), [], tuple()]
)
def test_validate_parameters_no_data(data: LookupTableData) -> None:
    with pytest.raises(ValueError, match="supply some data"):
        validate_build_table_parameters(data, [], [], [])


@pytest.mark.parametrize(
    "key_cols, param_cols, val_cols, match",
    [
        ((), (), (), "supply value_columns"),
        ((), (), [], "supply value_columns"),
        ((), (), ["a", "b"], "match the number of values"),
        (("a", "b"), (), ["d", "e", "f"], "key_columns are not allowed"),
        ((), ("a", "b"), ["d", "e", "f"], "parameter_columns are not allowed"),
    ],
)
def test_validate_parameters_error_scalar_data(
    key_cols: Sequence[str], param_cols: Sequence[str], val_cols: Sequence[str], match: str
) -> None:
    with pytest.raises(ValueError, match=match):
        validate_build_table_parameters([1, 2, 3], key_cols, param_cols, val_cols)


@pytest.mark.parametrize("data", ["FAIL", pd.Interval(5, 10), "2019-05-17"])
def test_validate_parameters_fail_other_data(data: LookupTableData) -> None:
    with pytest.raises(TypeError, match="only allowable types"):
        validate_build_table_parameters(data, [], [], [])


@pytest.mark.parametrize(
    "key_cols, param_cols, val_cols, match",
    [
        ([], [], ["c"], "either key_columns or parameter_columns"),
        (["a", "b"], ["b"], ["c"], "no overlap between key.*and parameter columns"),
        (["a"], ["b"], ["a", "c"], "no overlap between value.*and key.*columns"),
        (["a"], ["b"], ["b", "c"], "no overlap between value.*and.*parameter columns"),
        (["d"], ["b"], ["c"], "columns.*must all be present"),
        (["a"], ["d"], ["c"], "columns.*must all be present"),
        (["a"], ["b"], ["d"], "columns.*must all be present"),
    ],
)
def test_validate_parameters_error_dataframe(
    key_cols: Sequence[str], param_cols: Sequence[str], val_cols: Sequence[str], match: str
) -> None:
    data = pd.DataFrame({"a": [1, 2], "b_start": [0, 5], "b_end": [5, 10], "c": [100, 150]})
    with pytest.raises(ValueError, match=match):
        validate_build_table_parameters(data, key_cols, param_cols, val_cols)


def test_validate_parameters_pass_scalar_data() -> None:
    validate_build_table_parameters([1, 2, 3], (), (), ["a", "b", "c"])


@pytest.mark.parametrize(
    "key_cols, param_cols, val_cols",
    [
        (["a"], ["b"], ["c"]),
        ([], ["b"], ["c", "a"]),
        ([], ["b"], ["a", "c"]),
        ([], ["b"], ["c"]),
        (["a"], [], ["c"]),
        (["a"], ["b"], []),
        (["a"], [], []),
        ([], ["b"], []),
    ],
)
def test_validate_parameters_pass_dataframe(
    key_cols: Sequence[str], param_cols: Sequence[str], val_cols: Sequence[str]
) -> None:
    data = pd.DataFrame({"a": [1, 2], "b_start": [0, 5], "b_end": [5, 10], "c": [100, 150]})
    validate_build_table_parameters(data, key_cols, param_cols, val_cols)


@pytest.mark.parametrize("validate", [True, False])
def test_validate_flag(mocker: MockerFixture, validate: bool) -> None:
    manager = LookupTableManager()
    manager.setup(mocker.Mock())
    manager._validate = validate
    interface = LookupTableInterface(manager)

    mock_validator = mocker.patch(
        "vivarium.framework.lookup.manager.validate_build_table_parameters"
    )

    interface.build_table(0, value_columns=["a"])

    if validate:
        mock_validator.assert_called_once()
    else:
        mock_validator.assert_not_called()


def test__build_table_from_dict(base_config: LayeredConfigTree) -> None:
    simulation = InteractiveContext(components=[TestPopulation()], configuration=base_config)
    manager = simulation._tables
    data = {
        "a_start": [0.0, 0.5, 1.0, 1.5],
        "a_end": [0.5, 1.0, 1.5, 2.0],
        "b": [10.0, 20.0, 30.0, 40.0],
        "c": [100.0, 200.0, 300.0, 400.0],
    }
    # We convert the dict to a dataframe before we call validate_build_table_parameters so
    # this test is really going to just ensure we don't error out when we pass in a dict and
    # we get the expected return type from _build_table
    table = manager._build_table(
        data,  # type: ignore [arg-type]
        key_columns=["b"],
        parameter_columns=["a"],
        value_columns=["c"],
    )
    assert isinstance(table, InterpolatedTable)
    assert table.key_columns == ["b"]
    assert table.parameter_columns == ["a"]
    assert table.value_columns == ["c"]



import numpy

numpy.seterr(all="raise")

from vivarium.__about__ import (
    __author__,
    __copyright__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __uri__,
)
from vivarium._version import __version__
from vivarium.component import Component
from vivarium.framework.artifact import Artifact
from vivarium.framework.configuration import build_model_specification
from vivarium.framework.results.observer import Observer
from vivarium.interface import InteractiveContext



"""
=================
Logging Utilities
=================

This module contains utilities for configuring logging.

"""
from __future__ import annotations

import logging
import sys
from pathlib import Path
from typing import TYPE_CHECKING, TextIO

from loguru import logger


def configure_logging_to_terminal(verbosity: int, long_format: bool = True) -> None:
    """Configure logging to print to the sys.stdout.

    Parameters
    ----------
    verbosity
        The verbosity level of the logging. 0 logs at the WARNING level, 1 logs
        at the INFO level, and 2 logs at the DEBUG level.
    long_format
        Whether to use the long format for logging messages, which includes explicit
        information about the simulation context and component in the log messages.
    """
    _clear_default_configuration()
    _add_logging_sink(
        sink=sys.stdout,
        verbosity=verbosity,
        long_format=long_format,
        colorize=True,
        serialize=False,
    )


def configure_logging_to_file(output_directory: Path) -> None:
    """Configure logging to write to a file in the provided output directory.

    Parameters
    ----------
    output_directory
        The directory to write the log file to.
    """
    log_file = output_directory / "simulation.log"
    _add_logging_sink(
        log_file,
        verbosity=2,
        long_format=True,
        colorize=False,
        serialize=False,
    )


def _clear_default_configuration() -> None:
    try:
        logger.remove(0)  # Clear default configuration
    except ValueError:
        pass


def _add_logging_sink(
    sink: Path | TextIO,
    verbosity: int,
    long_format: bool,
    colorize: bool,
    serialize: bool,
) -> int:
    """Add a logging sink to the logger.

    Parameters
    ----------
    sink
        The sink to add.  Can be a file path, a file object, or a callable.
    verbosity
        The verbosity level.  0 is the default and will only log warnings and errors.
        1 will log info messages.  2 will log debug messages.
    long_format
        Whether to use the long format for logging messages.  The long format includes
        the simulation name and component name.  The short format only includes the
        file name and line number.
    colorize
        Whether to colorize the log messages.
    serialize
        Whether to serialize log messages.  This is useful when logging to
        a file or a database.
    """
    log_formatter = _LogFormatter(long_format)
    logging_level = _get_log_level(verbosity)
    return logger.add(
        sink,
        colorize=colorize,
        level=logging_level,
        format=log_formatter.format,
        serialize=serialize,
    )


class _LogFormatter:
    time = "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green>"
    level = "<level>{level: <8}</level>"
    simulation = "<cyan>{extra[simulation]}</cyan> - <cyan>{name}</cyan>:<cyan>{line}</cyan>"
    simulation_and_component = (
        "<cyan>{extra[simulation]}</cyan>-<cyan>{extra[component]}</cyan>:<cyan>{line}</cyan>"
    )
    short_name_and_line = "<cyan>{name}</cyan>:<cyan>{line}</cyan>"
    message = "<level>{message}</level>"

    def __init__(self, long_format: bool = False):
        self.long_format = long_format

    if TYPE_CHECKING:
        from loguru import Record

    def format(self, record: Record) -> str:
        fmt = self.time + " | "

        if self.long_format:
            fmt += self.level + " | "

        if self.long_format and "simulation" in record["extra"]:
            if "component" in record["extra"]:
                fmt += self.simulation_and_component + " - "
            else:
                fmt += self.simulation + " - "
        else:
            fmt += self.short_name_and_line + " - "

        fmt += self.message + "\n{exception}"
        return fmt


def _get_log_level(verbosity: int) -> str:
    if verbosity == 0:
        return "WARNING"
    elif verbosity == 1:
        return "INFO"
    elif verbosity >= 2:
        return "DEBUG"
    else:
        raise ValueError(f"Invalid verbosity level: {verbosity}")


def list_loggers() -> None:
    """Utility function for analyzing the logging environment."""
    root_logger = logging.getLogger()
    print("Root logger: ", root_logger)
    for h in root_logger.handlers:
        print(f"     %s" % h)

    print("Other loggers")
    print("=============")
    for name, logger_ in logging.Logger.manager.loggerDict.items():
        print("+ [%-20s] %s " % (name, logger_))
        if not isinstance(logger_, logging.PlaceHolder):
            handlers = list(logger_.handlers)
            if not handlers:
                print("     No handlers")
            for h in logger_.handlers:
                print("     %s" % h)



import pandas as pd
import pytest
from layered_config_tree import LayeredConfigTree

from tests.helpers import ColumnCreator, ColumnRequirer
from vivarium import InteractiveContext
from vivarium.framework.randomness.index_map import IndexMap
from vivarium.framework.randomness.manager import RandomnessError, RandomnessManager
from vivarium.framework.randomness.stream import get_hash


def mock_clock() -> pd.Timestamp:
    return pd.Timestamp("1/1/2005")


def test_randomness_manager_get_randomness_stream() -> None:
    seed = "123456"
    component = ColumnCreator()

    rm = RandomnessManager()
    rm._add_constraint = lambda f, **kwargs: f
    rm._seed = seed
    rm._clock_ = mock_clock
    rm._key_columns = ["age", "sex"]
    rm._key_mapping_ = IndexMap(["age", "sex"])
    stream = rm._get_randomness_stream("test", component)

    assert stream.key == "test"
    assert stream.seed == seed
    assert stream.clock is mock_clock
    assert set(rm._decision_points.keys()) == {"test"}
    assert stream.component == component

    with pytest.raises(RandomnessError):
        rm._get_randomness_stream("test", ColumnRequirer())


def test_randomness_manager_register_simulants() -> None:
    seed = "123456"
    rm = RandomnessManager()
    rm._add_constraint = lambda f, **kwargs: f
    rm._seed = seed
    rm._clock_ = mock_clock
    rm._key_columns = ["age", "sex"]
    rm._key_mapping_ = IndexMap(["age", "sex"])

    bad_df = pd.DataFrame({"age": range(10), "not_sex": [1] * 5 + [2] * 5})
    with pytest.raises(RandomnessError):
        rm.register_simulants(bad_df)

    good_df = pd.DataFrame({"age": range(10), "sex": [1] * 5 + [2] * 5})
    rm.register_simulants(good_df)

    assert isinstance(rm._key_mapping._map, pd.Series)
    map_index = rm._key_mapping._map.droplevel(rm._key_mapping.SIM_INDEX_COLUMN).index
    good_index = good_df.set_index(good_df.columns.tolist()).index
    assert map_index.difference(good_index).empty


def test_get_random_seed() -> None:
    seed = "123456"
    decision_point = "test"

    rm = RandomnessManager()
    rm._add_constraint = lambda f, **kwargs: f
    rm._seed = seed
    rm._clock_ = mock_clock

    assert rm.get_seed(decision_point) == get_hash(f"{decision_point}_{rm._clock()}_{seed}")


@pytest.mark.parametrize("additional_seed", ["789", None])
def test_additional_seed(base_config: LayeredConfigTree, additional_seed: str | None) -> None:

    input_draw = "123"
    seed = "456"
    base_config.update(
        {
            "input_data": {
                "input_draw_number": input_draw,
            },
            "randomness": {
                "random_seed": seed,
                "additional_seed": additional_seed,
            },
        },
    )

    component = ColumnCreator()
    sim = InteractiveContext(components=[component], configuration=base_config)
    rm = sim._randomness

    if additional_seed is not None:
        expected = f"{seed}_{additional_seed}"
    else:
        expected = f"{seed}_{input_draw}"
    assert rm._seed == expected



# mypy: ignore-errors
from __future__ import annotations

from typing import Any

import pandas as pd

from vivarium import Component, Observer
from vivarium.framework.engine import Builder
from vivarium.framework.event import Event
from vivarium.framework.population import SimulantData
from vivarium.framework.resource import Resource
from vivarium.manager import Manager


class MockComponentA(Observer):
    @property
    def name(self) -> str:
        return self._name

    @property
    def configuration_defaults(self) -> dict[str, Any]:
        return {}

    def __init__(self, *args, name: str = "mock_component_a") -> None:
        super().__init__()
        self._name = name
        self.args = args
        self.builder_used_for_setup = None

    def create_lookup_tables(self, builder: Builder) -> dict[str, Any]:
        return {}

    def register_observations(self, builder: Builder) -> None:
        pass

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.name == other.name


class MockComponentB(Observer):
    @property
    def name(self) -> str:
        return self._name

    def __init__(self, *args, name: str = "mock_component_b") -> None:
        super().__init__()
        self._name = name
        self.args = args
        self.builder_used_for_setup = None
        if len(self.args) > 1:
            for arg in self.args:
                self._sub_components.append(MockComponentB(arg, name=arg))

    def setup(self, builder: Builder) -> None:
        self.builder_used_for_setup = builder

    def register_observations(self, builder: Builder) -> None:
        builder.results.register_adding_observation(self.name, aggregator=self.counter)

    def create_lookup_tables(self, builder: Builder) -> dict[str, Any]:
        return {}

    def counter(self, _: Any) -> float:
        return 1.0

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.name == other.name


class MockGenericComponent(Component):
    CONFIGURATION_DEFAULTS = {
        "component": {
            "key1": "val",
            "key2": {
                "subkey1": "val",
                "subkey2": "val",
            },
            "key3": ["val", "val", "val"],
        }
    }

    @property
    def name(self) -> str:
        return self._name

    @property
    def configuration_defaults(self) -> dict[str, Any]:
        return {self.name: self.CONFIGURATION_DEFAULTS["component"]}

    def __init__(self, name: str):
        super().__init__()
        self._name = name
        self.builder_used_for_setup = None

    def setup(self, builder: Builder) -> None:
        self.builder_used_for_setup = builder
        self.config = builder.configuration[self.name]

    def __eq__(self, other: Any) -> bool:
        return type(self) == type(other) and self.name == other.name


class Listener(MockComponentB):
    def __init__(self, *args, name="test_listener"):
        super().__init__(*args, name=name)
        self.post_setup_called = False
        self.time_step_prepare_called = False
        self.time_step_called = False
        self.time_step_cleanup_called = False
        self.collect_metrics_called = False
        self.simulation_end_called = False

        self.event_indexes = {
            "time_step_prepare": None,
            "time_step": None,
            "time_step_cleanup": None,
            "collect_metrics": None,
        }

    def on_post_setup(self, event: Event) -> None:
        self.post_setup_called = True

    def on_time_step_prepare(self, event: Event) -> None:
        self.time_step_prepare_called = True
        self.event_indexes["time_step_prepare"] = event.index

    def on_time_step(self, event: Event) -> None:
        self.time_step_called = True
        self.event_indexes["time_step"] = event.index

    def on_time_step_cleanup(self, event: Event) -> None:
        self.time_step_cleanup_called = True
        self.event_indexes["time_step_cleanup"] = event.index

    def on_collect_metrics(self, event: Event) -> None:
        self.collect_metrics_called = True
        self.event_indexes["collect_metrics"] = event.index

    def on_simulation_end(self, event: Event) -> None:
        self.simulation_end_called = True


class ColumnCreator(Component):
    @property
    def columns_created(self) -> list[str]:
        return ["test_column_1", "test_column_2", "test_column_3"]

    def setup(self, builder: Builder) -> None:
        builder.value.register_value_producer("pipeline_1", lambda x: x)

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        self.population_view.update(self.get_initial_state(pop_data.index))

    def get_initial_state(self, index: pd.Index[int]) -> pd.DataFrame:
        return pd.DataFrame(
            {column: [i % 3 for i in index] for column in self.columns_created}, index=index
        )


class LookupCreator(ColumnCreator):
    CONFIGURATION_DEFAULTS = {
        "lookup_creator": {
            "data_sources": {
                "favorite_team": "simulants.favorite_team",
                "favorite_scalar": 0.4,
                "favorite_color": "simulants.favorite_color",
                "favorite_number": "simulants.favorite_number",
                "baking_time": "self::load_baking_time",
                "cooling_time": "tests.framework.components.test_component::load_cooling_time",
            },
            "alternate_data_sources": {
                "favorite_list": [9, 4],
            },
        }
    }

    def build_all_lookup_tables(self, builder: "Builder") -> None:
        super().build_all_lookup_tables(builder)
        self.lookup_tables["favorite_list"] = self.build_lookup_table(
            builder,
            self.configuration["alternate_data_sources"]["favorite_list"],
            ["column_1", "column_2"],
        )

    @staticmethod
    def load_baking_time(_builder: Builder) -> float:
        return 0.5


class SingleLookupCreator(ColumnCreator):
    pass


class OrderedColumnsLookupCreator(Component):
    @property
    def columns_created(self) -> list[str]:
        return ["foo", "bar"]

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        initialization_data = pd.DataFrame(
            {
                "foo": "key1",
                "bar": 15,
            },
            index=pop_data.index,
        )
        self.population_view.update(initialization_data)

    def build_all_lookup_tables(self, builder: "Builder") -> None:
        value_columns = ["one", "two", "three", "four", "five", "six", "seven"]
        ordered_columns = pd.DataFrame(
            [[1, 2, 3, 4, 5, 6, 7], [8, 9, 10, 11, 12, 13, 14]],
            columns=value_columns,
        )
        ordered_columns_categorical = ordered_columns.copy()
        ordered_columns_categorical["foo"] = ["key1", "key2"]
        ordered_columns_interpolated = ordered_columns.copy()
        ordered_columns_interpolated["foo"] = ["key1", "key1"]
        ordered_columns_interpolated["bar_start"] = [10, 20]
        ordered_columns_interpolated["bar_end"] = [20, 30]
        self.lookup_tables["ordered_columns_categorical"] = self.build_lookup_table(
            builder,
            ordered_columns_categorical,
            value_columns,
        )
        self.lookup_tables["ordered_columns_interpolated"] = self.build_lookup_table(
            builder,
            ordered_columns_interpolated,
            value_columns,
        )


class ColumnRequirer(Component):
    @property
    def columns_required(self) -> list[str]:
        return ["test_column_1", "test_column_2"]


class ColumnCreatorAndRequirer(Component):
    @property
    def columns_required(self) -> list[str]:
        return ["test_column_1", "test_column_2"]

    @property
    def columns_created(self) -> list[str]:
        return ["test_column_4"]

    @property
    def initialization_requirements(self) -> list[str | Resource]:
        return ["test_column_2", self.pipeline, self.randomness]

    def setup(self, builder: Builder) -> None:
        self.pipeline = builder.value.get_value("pipeline_1")
        self.randomness = builder.randomness.get_stream("stream_1")

    def on_initialize_simulants(self, pop_data: SimulantData) -> None:
        initialization_data = pd.DataFrame({"test_column_4": 8}, index=pop_data.index)
        self.population_view.update(initialization_data)


class ColumnCreatorAndAllRequirer(ColumnCreatorAndRequirer):
    @property
    def columns_required(self) -> list[str]:
        return []


class AllColumnsRequirer(Component):
    @property
    def columns_required(self) -> list[str]:
        return []


class FilteredPopulationView(ColumnRequirer):
    @property
    def population_view_query(self) -> str | None:
        return "test_column_1 == 5"


class NoPopulationView(Component):
    pass


class Parameterized(Component):
    def __init__(self, p_one: str, p_two: int, p_three: str):
        super().__init__()
        self.p_one = p_one
        self.p_two = p_two
        self._p_three = p_three


class ParameterizedByComponent(Component):
    def __init__(self, param: Parameterized):
        super().__init__()
        self.param = param


class DefaultPriorities(Component):
    def on_post_setup(self, event: Event) -> None:
        pass

    def on_time_step_prepare(self, event: Event) -> None:
        pass

    def on_time_step(self, event: Event) -> None:
        pass

    def on_time_step_cleanup(self, event: Event) -> None:
        pass

    def on_collect_metrics(self, event: Event) -> None:
        pass

    def on_simulation_end(self, event: Event) -> None:
        pass


class CustomPriorities(DefaultPriorities):
    @property
    def post_setup_priority(self) -> int:
        return 8

    @property
    def time_step_prepare_priority(self) -> int:
        return 7

    @property
    def time_step_priority(self) -> int:
        return 2

    @property
    def time_step_cleanup_priority(self) -> int:
        return 3

    @property
    def collect_metrics_priority(self) -> int:
        return 6

    @property
    def simulation_end_priority(self) -> int:
        return 1


class MockManager(Manager):
    @property
    def name(self) -> str:
        return self._name

    def __init__(self, name: str) -> None:
        self._name = name
